import { Inject, Injectable } from '@nestjs/common';
import { ClientProxy } from '@nestjs/microservices';
import { interval, Observable, Subject } from 'rxjs';
import * as net from 'net'
import { ConfigService } from '@nestjs/config';
import { connectToAfisJava } from 'libs/java/afis.utils';
import { Fingerprint, FingerprintPayload, FingerprintPayloadUI, FisMessage, JavaResponse, PersonFingerprintData } from 'libs';
import { v1 as uuid } from 'uuid'

@Injectable()
export class FisVerificationService {
  private port: number
  private host: string
  private javaClient: net.Socket
  private incomingMessageFromJava!: Subject<JavaResponse>
  private registeredFingerprintData: Fingerprint[] = []

  constructor(
    @Inject(`SAMPLEAPP_SERVICE`) private sampleClient: ClientProxy,
    @Inject(`FINGERPRINT_SERVICE`) private fingerprintClient: ClientProxy,
    private configService: ConfigService
  ) {
    // Set up incoming responses or messages from TCP Java
    this.port = this.configService.get<number>('afis.tcpPort') as number
    this.host = this.configService.get<string>('afis.host') as string

    this.javaClient = connectToAfisJava(this.host, this.port)
    this.javaClient.on(`data`, (data) => {
      let message: JavaResponse = JSON.parse(data.toString())
      console.log(message)
      console.log(`Java server response: ${message.message}`)
      if (message.operation == `QualityAssurance`) {
        let response: FisMessage = {
          header: {
            messageID: uuid(),
            messageName: `Verification Response`
          },
          data: {
            id: uuid(),
            cmd: `QualityAssurance`,
            message: `${message.data.name} finger position ${message.data.fingerprints[0].fpPosition} edge score is ${(message.edgeScore > 250 ? `good` : `bad`)}`,
            data: {
              score: message.score,
              person: message.data.name,
              fpPosition: message.data.fingerprints[0].fpPosition,
              fpQuality: (message.score > 250 ? `good` : `bad`)
            }
          }
        }
        this.fingerprintClient.emit(`message`, response)
      }

      if (message.operation == `Registration` && message.status == `Success`) {
        this.registeredFingerprintData.push((message.data as PersonFingerprintData).fingerprints[0] as Fingerprint)
        console.log(message.data)
        // console.log(this.registeredFingerprintData)
        console.log(`${this.registeredFingerprintData.length} registered print.`)
      }
      // this.incomingMessageFromJava.next(message)
    })
  }

  public handleMessage(message: FisMessage): void {
    console.log(`Sending fingerprint data to Java verification....`)
    let payload: FingerprintPayloadUI = message.data as FingerprintPayloadUI
    let command: FingerprintPayload = {
      ...payload,
      fpTemplateArray: this.registeredFingerprintData
    }
    this.javaClient.write(JSON.stringify(command) + `\n`)
    // let data = JSON.parse(message)
    // console.log(data)
    // this.javaClient.write(JSON.stringify(extractionMessage) + `\n`)
  }

}

/* Reason based on ChatGPT: Why ClientProxy.send() or emit() won't work 
The NestJS ClientProxy (when configured for Transport.TCP) uses its own custom binary protocol under the hood. This includes:
A message envelope that wraps your pattern and data.
Message IDs and event headers.
A specific serialization/deserialization logic that is not plain JSON.
NO newline character at the end.
So when your Java server, which expects a newline-terminated JSON message, receives this custom binary format, it gets confused or sees null.*/

/* data: {
messageData: string
messageTemplateData: string
deviceInfo: object} */